@@ -3,12 +3,16 @@
  * in the RFC. See iscsi_proto.h for those.
  */
 /* timeouts in seconds */
-#define DEF_LOGIN_TIMEO		15
+#define DEF_LOGIN_TIMEO		30
 #define DEF_LOGOUT_TIMEO	15
-#define DEF_NOOP_OUT_INTERVAL	10
-#define DEF_NOOP_OUT_TIMEO	15
+#define DEF_NOOP_OUT_INTERVAL	5
+#define DEF_NOOP_OUT_TIMEO	5
 #define DEF_REPLACEMENT_TIMEO	120
 
+#define DEF_ABORT_TIMEO		15
+#define DEF_LU_RESET_TIMEO	30
+#define DEF_HOST_RESET_TIMEO	60
+
 /* q depths */
 #define CMDS_MAX	128
 #define QUEUE_DEPTH	32
