@@ -42,6 +42,7 @@
 #include "version.h"
 #include "iscsi_sysfs.h"
 #include "iscsi_settings.h"
+#include "fw_context.h"
 
 /* global config info */
 /* initiator needs initiator name/alias */
@@ -68,6 +69,8 @@ static struct option const long_options[] = {
 	{"username_in", required_argument, NULL, 'U'},
 	{"password_in", required_argument, NULL, 'W'},
 	{"debug", required_argument, NULL, 'd'},
+	{"fwparam_connect", no_argument, NULL, 'b'},
+	{"fwparam_print", no_argument, NULL, 'f'},
 	{"help", no_argument, NULL, 'h'},
 	{"version", no_argument, NULL, 'v'},
 	{NULL, 0, NULL, 0},
@@ -92,6 +95,8 @@ Open-iSCSI initiator.\n\
   -U, --username_in=N      set incoming username to N (optional)\n\
   -W, --password_in=N      set incoming password to N (optional\n\
   -d, --debug debuglevel   print debugging information \n\
+  -b, --fwparam_connect    create a session to the target\n\
+  -f, --fwparam_print      print the iBFT to STDOUT \n\
   -h, --help               display this help and exit\n\
   -v, --version            display version and exit\n\
 ");
@@ -108,11 +113,10 @@ static int stop_event_loop(void)
 	memset(&req, 0, sizeof(req));
 	req.command = MGMT_IPC_IMMEDIATE_STOP;
 	rc = do_iscsid(&req, &rsp);
-	if (rc)
+	if (rc) {
 		iscsid_handle_error(rc);
-	if (rc)
 		log_error("Could not stop event_loop\n");
-
+	}
 	return rc;
 }
 
@@ -122,10 +126,19 @@ static int setup_session(void)
 	iscsiadm_rsp_t rsp;
 	int rc;
 
+	/*
+	 * For root boot we cannot change this so increase to account
+	 * for boot using static setup.
+	 */
+	config_rec.session.initial_login_retry_max = 30;
 	/* we cannot answer so turn off */
 	config_rec.conn[0].timeo.noop_out_interval = 0;
 	config_rec.conn[0].timeo.noop_out_timeout = 0;
 
+	printf("%s: Logging into %s %s:%d,%d\n", program_name, config_rec.name,
+		config_rec.conn[0].address, config_rec.conn[0].port,
+		config_rec.tpgt);
+
 	memset(&req, 0, sizeof(req));
 	req.command = MGMT_IPC_SESSION_LOGIN;
 	memcpy(&req.u.session.rec, &config_rec, sizeof(node_rec_t));
@@ -145,22 +158,22 @@ static int check_params(char *initiatorname)
 {
 	if (!initiatorname) {
 		log_error("InitiatorName not set. Exiting %s\n", program_name);
-		return -EINVAL;
+		return EINVAL;
 	}
 
 	if (config_rec.tpgt == PORTAL_GROUP_TAG_UNKNOWN) {
 		log_error("Portal Group not set. Exiting %s\n", program_name);
-		return -EINVAL;
+		return EINVAL;
 	}
 
 	if (!strlen(config_rec.name)) {
 		log_error("TargetName not set. Exiting %s\n", program_name);
-		return -EINVAL;
+		return EINVAL;
 	}
 
 	if (!strlen(config_rec.conn[0].address)) {
 		log_error("IP Address not set. Exiting %s\n", program_name);
-		return -EINVAL;
+		return EINVAL;
 	}
 
 	return 0;
@@ -175,15 +188,15 @@ do {									\
 	}								\
 } while (0);
 
-/* TODO fix leaks */
 int main(int argc, char *argv[])
 {
 	struct utsname host_info; /* will use to compound initiator alias */
 	struct iscsi_auth_config *auth;
 	char *initiatorname = NULL;
-	int ch, longindex;
+	int ch, longindex, ret;
 	struct sigaction sa_old;
 	struct sigaction sa_new;
+	struct boot_context context;
 	pid_t pid;
 
 	idbm_node_setup_defaults(&config_rec);
@@ -197,7 +210,7 @@ int main(int argc, char *argv[])
 	sa_new.sa_flags = 0;
 	sigaction(SIGINT, &sa_new, &sa_old );
 
-	while ((ch = getopt_long(argc, argv, "i:t:g:a:p:d:u:w:U:W:vh",
+	while ((ch = getopt_long(argc, argv, "i:t:g:a:p:d:u:w:U:W:bfvh",
 				 long_options, &longindex)) >= 0) {
 		switch (ch) {
 		case 'i':
@@ -246,6 +259,44 @@ int main(int argc, char *argv[])
 		case 'd':
 			log_level = atoi(optarg);
 			break;
+		case 'b':
+			ret = fw_get_entry(&context, NULL);
+			if (ret) {
+				printf("Could not setup fw entries.");
+				exit(-1);
+			}
+
+			initiatorname = context.initiatorname;
+			strncpy(config_rec.name, context.targetname,
+				sizeof(context.targetname));
+			strncpy(config_rec.conn[0].address,
+				context.target_ipaddr,
+				sizeof(context.target_ipaddr));
+			config_rec.conn[0].port = context.target_port;
+			/* this seems broken ??? */
+			config_rec.tpgt = 1;
+			strncpy(auth->username, context.chap_name,
+				sizeof(context.chap_name));
+			strncpy((char *)auth->password, context.chap_password,
+				sizeof(context.chap_password));
+			auth->password_length = strlen((char *)auth->password);
+			strncpy(auth->username_in, context.chap_name_in,
+				sizeof(context.chap_name_in));
+			strncpy((char *)auth->password_in,
+				context.chap_password_in,
+				sizeof(context.chap_password_in));
+			auth->password_in_length =
+					strlen((char *)auth->password_in);
+			break;
+		case 'f':
+			ret = fw_get_entry(&context, NULL);
+			if (ret) {
+				printf("Could not read fw values.\n");
+				exit(-1);
+			}
+
+			fw_print_entry(&context);
+			exit(0);
 		case 'v':
 			printf("%s version %s\n", program_name,
 				ISCSI_VERSION_STR);
@@ -263,7 +314,10 @@ int main(int argc, char *argv[])
 	log_daemon = 0;
 	log_init(program_name, DEFAULT_AREA_SIZE);
 
-	check_class_version();
+	sysfs_init();
+	if (iscsi_sysfs_check_class_version())
+		exit(1);
+
 	if (check_params(initiatorname))
 		exit(1);
 
@@ -334,6 +388,7 @@ int main(int argc, char *argv[])
 	ipc->ctldev_close();
 	mgmt_ipc_close(mgmt_ipc_fd);
 	free_initiator();
+	sysfs_cleanup();
 
 	log_debug(1, "iscsi child done");
 	return 0;
