@@ -46,6 +46,7 @@ typedef enum mgmt_ipc_err {
 	MGMT_IPC_ERR_INVALID_REQ	= 16,
 	MGMT_IPC_ERR_ISNS_UNAVAILABLE	= 17,
 	MGMT_IPC_ERR_ISCSID_COMM_ERR	= 18,
+	MGMT_IPC_ERR_FATAL_LOGIN_FAILURE = 19,
 } mgmt_ipc_err_e;
 
 typedef enum iscsiadm_cmd {
@@ -65,28 +66,18 @@ typedef enum iscsiadm_cmd {
 	MGMT_IPC_ISNS_DEV_ATTR_QUERY	= 14,
 	MGMT_IPC_SEND_TARGETS		= 15,
 	MGMT_IPC_SET_HOST_PARAM		= 16,
+	MGMT_IPC_NOTIFY_ADD_NODE	= 17,
+	MGMT_IPC_NOTIFY_DEL_NODE	= 18,
+	MGMT_IPC_NOTIFY_ADD_PORTAL	= 19,
+	MGMT_IPC_NOTIFY_DEL_PORTAL	= 20,
+
+	__MGMT_IPC_MAX_COMMAND
 } iscsiadm_cmd_e;
 
-typedef enum iscsi_conn_state_e {
-	STATE_FREE,
-	STATE_XPT_WAIT,
-	STATE_IN_LOGIN,
-	STATE_LOGGED_IN,
-	STATE_IN_LOGOUT,
-	STATE_LOGOUT_REQUESTED,
-	STATE_CLEANUP_WAIT,
-} iscsi_conn_state_e;
-
-typedef enum iscsi_session_r_stage_e {
-	R_STAGE_NO_CHANGE,
-	R_STAGE_SESSION_CLEANUP,
-	R_STAGE_SESSION_REOPEN,
-	R_STAGE_SESSION_REDIRECT,
-} iscsi_session_r_stage_e;
-
 /* IPC Request */
 typedef struct iscsiadm_req {
 	iscsiadm_cmd_e command;
+	uint32_t payload_len;
 
 	union {
 		/* messages */
@@ -133,12 +124,15 @@ typedef struct iscsiadm_rsp {
 			char var[VALUE_MAXLEN];
 		} config;
 		struct ipc_msg_session_state {
-			iscsi_session_r_stage_e session_state;
-			iscsi_conn_state_e conn_state;
+			int session_state;
+			int conn_state;
 		} session_state;
 	} u;
 } iscsiadm_rsp_t;
 
+struct queue_task;
+typedef mgmt_ipc_err_e	mgmt_ipc_fn_t(struct queue_task *);
+
 struct iscsi_ipc *ipc;
 
 void need_reap(void);
