@@ -25,6 +25,7 @@
 #include <unistd.h>
 #include <stdint.h>
 #include <errno.h>
+#include <inttypes.h>
 #include <asm/types.h>
 #include <sys/socket.h>
 #include <sys/types.h>
@@ -279,7 +280,18 @@ __kipc_call(void *iov_base, int iov_len)
 							 sizeof(*ev), 0)) < 0) {
 					return rc;
 				}
-				log_error("received iferror %d", ev->iferror);
+				if (ev->iferror == -ENOSYS)
+					/* not fatal so let caller handle log */
+					log_debug(1, "Recieved iferror %d: %s",
+						  ev->iferror,
+						  strerror(ev->iferror));
+				else if (ev->iferror < 0)
+					log_error("Received iferror %d: %s",
+						   ev->iferror,
+						   strerror(ev->iferror));
+				else
+					log_error("Received iferror %d",
+						   ev->iferror);
 				return ev->iferror;
 			}
 			/*
@@ -339,9 +351,9 @@ ksendtargets(uint64_t transport_handle, uint32_t host_no, struct sockaddr *addr)
 }
 
 static int
-kcreate_session(uint64_t transport_handle, uint32_t initial_cmdsn,
-		uint16_t cmds_max, uint16_t qdepth,
-		uint32_t *out_sid, uint32_t *out_hostno)
+kcreate_session(uint64_t transport_handle, uint64_t ep_handle,
+		uint32_t initial_cmdsn, uint16_t cmds_max, uint16_t qdepth,
+		uint32_t *out_sid, uint32_t *hostno)
 {
 	int rc;
 	struct iscsi_uevent ev;
@@ -350,17 +362,26 @@ kcreate_session(uint64_t transport_handle, uint32_t initial_cmdsn,
 
 	memset(&ev, 0, sizeof(struct iscsi_uevent));
 
-	ev.type = ISCSI_UEVENT_CREATE_SESSION;
-	ev.transport_handle = transport_handle;
-	ev.u.c_session.initial_cmdsn = initial_cmdsn;
-	ev.u.c_session.cmds_max = cmds_max;
-	ev.u.c_session.queue_depth = qdepth;
+	if (ep_handle == 0) {
+		ev.type = ISCSI_UEVENT_CREATE_SESSION;
+		ev.transport_handle = transport_handle;
+		ev.u.c_session.initial_cmdsn = initial_cmdsn;
+		ev.u.c_session.cmds_max = cmds_max;
+		ev.u.c_session.queue_depth = qdepth;
+	} else {
+		ev.type = ISCSI_UEVENT_CREATE_BOUND_SESSION;
+		ev.transport_handle = transport_handle;
+		ev.u.c_bound_session.initial_cmdsn = initial_cmdsn;
+		ev.u.c_bound_session.cmds_max = cmds_max;
+		ev.u.c_bound_session.queue_depth = qdepth;
+		ev.u.c_bound_session.ep_handle = ep_handle;
+	}
 
 	if ((rc = __kipc_call(&ev, sizeof(ev))) < 0) {
 		return rc;
 	}
 
-	*out_hostno = ev.r.c_session_ret.host_no;
+	*hostno = ev.r.c_session_ret.host_no;
 	*out_sid = ev.r.c_session_ret.sid;
 
 	return 0;
@@ -387,6 +408,27 @@ kdestroy_session(uint64_t transport_handle, uint32_t sid)
 	return 0;
 }
 
+static int
+kunbind_session(uint64_t transport_handle, uint32_t sid)
+{
+	int rc;
+	struct iscsi_uevent ev;
+
+	log_debug(7, "in %s", __FUNCTION__);
+
+	memset(&ev, 0, sizeof(struct iscsi_uevent));
+
+	ev.type = ISCSI_UEVENT_UNBIND_SESSION;
+	ev.transport_handle = transport_handle;
+	ev.u.d_session.sid = sid;
+
+	if ((rc = __kipc_call(&ev, sizeof(ev))) < 0) {
+		return rc;
+	}
+
+	return 0;
+}
+
 static int
 kcreate_conn(uint64_t transport_handle, uint32_t sid,
 	    uint32_t cid, uint32_t *out_cid)
@@ -668,8 +710,7 @@ kstart_conn(uint64_t transport_handle, uint32_t sid, uint32_t cid,
 }
 
 static int
-krecv_pdu_begin(uint64_t transport_handle, uintptr_t recv_handle,
-		uintptr_t *pdu_handle, int *pdu_size)
+krecv_pdu_begin(struct iscsi_conn *conn)
 {
 	log_debug(7, "in %s", __FUNCTION__);
 
@@ -677,19 +718,17 @@ krecv_pdu_begin(uint64_t transport_handle, uintptr_t recv_handle,
 		log_error("recv's begin state machine bug?");
 		return -EIO;
 	}
-	recvbuf = (void*)recv_handle + sizeof(struct iscsi_uevent);
+	recvbuf = conn->recv_context->data + sizeof(struct iscsi_uevent);
 	recvlen = 0;
-	*pdu_handle = recv_handle;
 
 	log_debug(3, "recv PDU began, pdu handle 0x%p",
-		  (void*)*pdu_handle);
+		  recvbuf);
 
 	return 0;
 }
 
 static int
-krecv_pdu_end(uint64_t transport_handle, uintptr_t conn_handle,
-	      uintptr_t pdu_handle)
+krecv_pdu_end(struct iscsi_conn *conn)
 {
 	log_debug(7, "in %s", __FUNCTION__);
 
@@ -699,9 +738,10 @@ krecv_pdu_end(uint64_t transport_handle, uintptr_t conn_handle,
 	}
 
 	log_debug(3, "recv PDU finished for pdu handle 0x%p",
-		  (void*)pdu_handle);
+		  recvbuf);
 
-	recvpool_put((void*)conn_handle, (void*)pdu_handle);
+	iscsi_conn_context_put(conn->recv_context);
+	conn->recv_context = NULL;
 	recvbuf = NULL;
 	return 0;
 }
@@ -774,7 +814,7 @@ ktransport_ep_disconnect(iscsi_conn_t *conn)
 
 	log_debug(7, "in %s", __FUNCTION__);
 
-	if (conn->transport_ep_handle < 0)
+	if (conn->transport_ep_handle == -1)
 		return;
 
 	memset(&ev, 0, sizeof(struct iscsi_uevent));
@@ -784,8 +824,9 @@ ktransport_ep_disconnect(iscsi_conn_t *conn)
 	ev.u.ep_disconnect.ep_handle = conn->transport_ep_handle;
 
 	if ((rc = __kipc_call(&ev, sizeof(ev))) < 0) {
-		log_error("conn %p session %p transport disconnect failed %d\n",
-			  conn, conn->session, rc);
+		log_error("connnection %d:%d transport disconnect failed for "
+			  "ep %" PRIu64 " with error %d.", conn->session->id,
+			  conn->id, conn->transport_ep_handle, rc);
 	} else
 		conn->transport_ep_handle = -1;
 }
@@ -847,15 +888,14 @@ static void drop_data(struct nlmsghdr *nlh)
 
 static int ctldev_handle(void)
 {
-	int rc;
+	int rc, ev_size;
 	struct iscsi_uevent *ev;
-	struct iscsi_transport *t;
 	iscsi_session_t *session = NULL;
 	iscsi_conn_t *conn = NULL;
-	uintptr_t recv_handle;
 	char nlm_ev[NLMSG_SPACE(sizeof(struct iscsi_uevent))];
 	struct nlmsghdr *nlh;
-	int ev_size;
+	struct iscsi_conn_context *conn_context;
+	uint32_t sid = 0, cid = 0;
 
 	log_debug(7, "in %s", __FUNCTION__);
 
@@ -870,6 +910,8 @@ static int ctldev_handle(void)
 	log_debug(7, "%s got event type %u\n", __FUNCTION__, ev->type);
 	/* drivers like qla4xxx can be inserted after iscsid is started */
 	switch (ev->type) {
+	case ISCSI_KEVENT_CREATE_SESSION:
+	/* old kernels sent ISCSI_UEVENT_CREATE_SESSION on creation */
 	case ISCSI_UEVENT_CREATE_SESSION:
 		drop_data(nlh);
 		iscsi_async_session_creation(ev->r.c_session_ret.host_no,
@@ -880,76 +922,72 @@ static int ctldev_handle(void)
 		iscsi_async_session_destruction(ev->r.d_session.host_no,
 						ev->r.d_session.sid);
 		return 0;
+	case ISCSI_KEVENT_RECV_PDU:
+		sid = ev->r.recv_req.sid;
+		cid = ev->r.recv_req.cid;
+		break;
+	case ISCSI_KEVENT_CONN_ERROR:
+		sid = ev->r.connerror.sid;
+		cid = ev->r.connerror.cid;
+	case ISCSI_KEVENT_UNBIND_SESSION:
+		sid = ev->r.unbind_session.sid;
+		/* session wide event so cid is 0 */
+		cid = 0;
+		break;
 	default:
 		; /* fall through */
 	}
 
 	/* verify connection */
-	list_for_each_entry(t, &transports, list) {
-		list_for_each_entry(session, &t->sessions, list) {
-			int i;
-
-			for (i=0; i<ISCSI_CONN_MAX; i++) {
-				if (ev->type == ISCSI_KEVENT_RECV_PDU &&
-				    session->id == ev->r.recv_req.sid &&
-				    session->conn[i].id == ev->r.recv_req.cid) {
-					conn = &session->conn[i];
-					goto verify_conn;
-				}
-				if (ev->type == ISCSI_KEVENT_CONN_ERROR &&
-				    session->id == ev->r.connerror.sid &&
-				    session->conn[i].id == ev->r.connerror.cid) {
-					conn = &session->conn[i];
-					goto verify_conn;
-				}
-			}
-		}
-	}
-
-verify_conn:
-	if (conn == NULL) {
+	session = session_find_by_sid(sid);
+	if (!session) {
 		log_error("Could not verify connection %d:%d. Dropping "
-			   "event.\n", ev->r.recv_req.sid, ev->r.recv_req.cid);
+			   "event.\n", sid, cid);
 		drop_data(nlh);
 		return -ENXIO;
 	}
+	conn = &session->conn[0];
 
 	ev_size = nlh->nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
-	recv_handle = (uintptr_t)recvpool_get(conn, ev_size);
-	if (!recv_handle) {
+	conn_context = iscsi_conn_context_get(conn, ev_size);
+	if (!conn_context) {
 		/* retry later */
-		log_error("Can not allocate memory for receive handle.");
+		log_error("Can not allocate memory for receive context.");
 		return -ENOMEM;
 	}
 
 	log_debug(6, "message real length is %d bytes, recv_handle %p",
-		nlh->nlmsg_len, (void*)recv_handle);
+		nlh->nlmsg_len, conn_context->data);
 
-	if ((rc = nlpayload_read(ctrl_fd, (void*)recv_handle,
+	if ((rc = nlpayload_read(ctrl_fd, conn_context->data,
 				ev_size, 0)) < 0) {
-		recvpool_put(conn, (void*)recv_handle);
+		iscsi_conn_context_put(conn_context);
 		log_error("can not read from NL socket, error %d", rc);
 		/* retry later */
 		return rc;
 	}
 
+	/*
+	 * we sched these events because the handlers could call back
+	 * into ctldev_handle
+	 */
 	switch (ev->type) {
 	case ISCSI_KEVENT_RECV_PDU:
-		/* produce an event, so session manager will handle */
-		queue_produce(session->queue, EV_CONN_RECV_PDU, conn,
-			sizeof(uintptr_t), &recv_handle);
-		actor_schedule(&session->mainloop);
+		iscsi_sched_conn_context(conn_context, conn, 0,
+					 EV_CONN_RECV_PDU);
 		break;
 	case ISCSI_KEVENT_CONN_ERROR:
-		/* produce an event, so session manager will handle */
-		memcpy((void *)recv_handle, &ev->r.connerror.error,
+		memcpy(conn_context->data, &ev->r.connerror.error,
 			sizeof(ev->r.connerror.error));
-		queue_produce(session->queue, EV_CONN_ERROR, conn,
-			sizeof(uintptr_t), &recv_handle);
-		actor_schedule(&session->mainloop);
+		iscsi_sched_conn_context(conn_context, conn, 0,
+					 EV_CONN_ERROR);
+		break;
+	case ISCSI_KEVENT_UNBIND_SESSION:
+		iscsi_sched_conn_context(conn_context, conn, 0,
+					 EV_CONN_STOP);
 		break;
 	default:
-		recvpool_put(conn, (void*)recv_handle);
+		iscsi_conn_context_put(conn_context);
 		log_error("unknown kernel event %d", ev->type);
 		return -EEXIST;
 	}
@@ -1028,12 +1066,12 @@ ctldev_close(void)
 {
 	log_debug(7, "in %s", __FUNCTION__);
 
+	if (ctrl_fd >= 0)
+		close(ctrl_fd);
 	free(setparam_buf);
 	free(pdu_sendbuf);
 	free(nlm_recvbuf);
 	free(nlm_sendbuf);
-	if (ctrl_fd >= 0)
-		close(ctrl_fd);
 }
 
 struct iscsi_ipc nl_ipc = {
@@ -1045,6 +1083,7 @@ struct iscsi_ipc nl_ipc = {
 	.sendtargets		= ksendtargets,
 	.create_session         = kcreate_session,
 	.destroy_session        = kdestroy_session,
+	.unbind_session		= kunbind_session,
 	.create_conn            = kcreate_conn,
 	.destroy_conn           = kdestroy_conn,
 	.bind_conn              = kbind_conn,
