@@ -2,15 +2,11 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include <ctype.h>
 #include <string.h>
 #include <errno.h>
-#include <pwd.h>
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <sys/un.h>
-#include <linux/types.h>
-#include <linux/unistd.h>
+#include <sys/time.h>
+#include <sys/resource.h>
 
 #include "log.h"
 #include "actor.h"
@@ -24,6 +20,7 @@
 #include "iscsi_proto.h"
 #include "transport.h"
 #include "idbm.h"
+#include "iface.h"
 
 void daemon_init(void)
 {
@@ -92,6 +89,35 @@ str_to_ipport(char *str, int *port, int *tpgt)
 	return ip;
 }
 
+#define ISCSI_MAX_FILES 16384
+
+int increase_max_files(void)
+{
+	struct rlimit rl;
+	int err;
+
+	err = getrlimit(RLIMIT_NOFILE, &rl);
+	if (err) {
+		log_debug(1, "Could not get file limit (err %d)\n", errno);
+		return errno;
+	}
+	log_debug(1, "Max file limits %lu %lu\n", rl.rlim_cur, rl.rlim_max);
+
+	if (rl.rlim_cur < ISCSI_MAX_FILES)
+		rl.rlim_cur = ISCSI_MAX_FILES;
+	if (rl.rlim_max < ISCSI_MAX_FILES)
+		rl.rlim_max = ISCSI_MAX_FILES;
+
+	err = setrlimit(RLIMIT_NOFILE, &rl);
+	if (err) {
+		log_debug(1, "Could not set file limit to %lu/%lu (err %d)\n",
+			  rl.rlim_cur, rl.rlim_max, errno);
+		return errno;
+	}
+
+	return 0;
+}
+
 #define MAXSLEEP 128
 
 static mgmt_ipc_err_e iscsid_connect(int *fd)
@@ -117,6 +143,11 @@ static mgmt_ipc_err_e iscsid_connect(int *fd)
 			/* Connection established */
 			return MGMT_IPC_OK;
 
+		/* If iscsid isn't there, there's no sense
+		 * in retrying. */
+		if (errno == ECONNREFUSED)
+			break;
+
 		/*
 		 * Delay before trying again
 		 */
@@ -127,27 +158,38 @@ static mgmt_ipc_err_e iscsid_connect(int *fd)
 	return MGMT_IPC_ERR_ISCSID_COMM_ERR;
 }
 
-static mgmt_ipc_err_e iscsid_request(int fd, iscsiadm_req_t *req)
+mgmt_ipc_err_e iscsid_request(int *fd, iscsiadm_req_t *req)
 {
 	int err;
 
-	if ((err = write(fd, req, sizeof(*req))) != sizeof(*req)) {
+	err = iscsid_connect(fd);
+	if (err)
+		return err;
+
+	if ((err = write(*fd, req, sizeof(*req))) != sizeof(*req)) {
 		log_error("got write error (%d/%d) on cmd %d, daemon died?",
 			err, errno, req->command);
+		close(*fd);
 		return MGMT_IPC_ERR_ISCSID_COMM_ERR;
 	}
 	return MGMT_IPC_OK;
 }
 
-static mgmt_ipc_err_e iscsid_response(int fd, iscsiadm_rsp_t *rsp)
+mgmt_ipc_err_e iscsid_response(int fd, iscsiadm_cmd_e cmd, iscsiadm_rsp_t *rsp)
 {
+	mgmt_ipc_err_e iscsi_err;
 	int err;
 
 	if ((err = recv(fd, rsp, sizeof(*rsp), MSG_WAITALL)) != sizeof(*rsp)) {
 		log_error("got read error (%d/%d), daemon died?", err, errno);
-		return MGMT_IPC_ERR_ISCSID_COMM_ERR;
+		iscsi_err = MGMT_IPC_ERR_ISCSID_COMM_ERR;
 	} else
-		return rsp->err;
+		iscsi_err = rsp->err;
+	close(fd);
+
+	if (!iscsi_err && cmd != rsp->command)
+		iscsi_err = MGMT_IPC_ERR_ISCSID_COMM_ERR;
+	return iscsi_err;
 }
 
 mgmt_ipc_err_e do_iscsid(iscsiadm_req_t *req, iscsiadm_rsp_t *rsp)
@@ -155,24 +197,61 @@ mgmt_ipc_err_e do_iscsid(iscsiadm_req_t *req, iscsiadm_rsp_t *rsp)
 	int fd;
 	mgmt_ipc_err_e err;
 
-	err = iscsid_connect(&fd);
+	err = iscsid_request(&fd, req);
 	if (err)
-		goto out;
+		return err;
 
-	err = iscsid_request(fd, req);
+	return iscsid_response(fd, req->command, rsp);
+}
+
+int iscsid_req_wait(iscsiadm_cmd_e cmd, int fd)
+{
+	iscsiadm_rsp_t rsp;
+
+	memset(&rsp, 0, sizeof(iscsiadm_rsp_t));
+	return iscsid_response(fd, cmd, &rsp);
+}
+
+int iscsid_req_by_rec_async(iscsiadm_cmd_e cmd, node_rec_t *rec, int *fd)
+{
+	iscsiadm_req_t req;
+
+	memset(&req, 0, sizeof(iscsiadm_req_t));
+	req.command = cmd;
+	memcpy(&req.u.session.rec, rec, sizeof(node_rec_t));
+
+	return iscsid_request(fd, &req);
+}
+
+int iscsid_req_by_rec(iscsiadm_cmd_e cmd, node_rec_t *rec)
+{
+	int err, fd;
+
+	err = iscsid_req_by_rec_async(cmd, rec, &fd);
 	if (err)
-		goto out;
+		return err;
+	return iscsid_req_wait(cmd, fd);
+}
 
-	err = iscsid_response(fd, rsp);
+int iscsid_req_by_sid_async(iscsiadm_cmd_e cmd, int sid, int *fd)
+{
+	iscsiadm_req_t req;
+
+	memset(&req, 0, sizeof(iscsiadm_req_t));
+	req.command = cmd;
+	req.u.session.sid = sid;
+
+	return iscsid_request(fd, &req);
+}
+
+int iscsid_req_by_sid(iscsiadm_cmd_e cmd, int sid)
+{
+	int err, fd;
+
+	err = iscsid_req_by_sid_async(cmd, sid, &fd);
 	if (err)
-		goto out;
-
-	if (!err && req->command != rsp->command)
-		err = MGMT_IPC_ERR_ISCSID_COMM_ERR;
-out:
-	if (fd >= 0)
-		close(fd);
-	return err;
+		return err;
+	return iscsid_req_wait(cmd, fd);
 }
 
 void idbm_node_setup_defaults(node_rec_t *rec)
@@ -191,43 +270,42 @@ void idbm_node_setup_defaults(node_rec_t *rec)
 	rec->session.auth.authmethod = 0;
 	rec->session.auth.password_length = 0;
 	rec->session.auth.password_in_length = 0;
-	rec->session.err_timeo.abort_timeout = 10;
-	rec->session.err_timeo.reset_timeout = 30;
+	rec->session.err_timeo.abort_timeout = DEF_ABORT_TIMEO;
+	rec->session.err_timeo.lu_reset_timeout = DEF_LU_RESET_TIMEO;
+	rec->session.err_timeo.host_reset_timeout = DEF_HOST_RESET_TIMEO;
 	rec->session.timeo.replacement_timeout = DEF_REPLACEMENT_TIMEO;
 	rec->session.iscsi.InitialR2T = 0;
 	rec->session.iscsi.ImmediateData = 1;
 	rec->session.iscsi.FirstBurstLength = DEF_INI_FIRST_BURST_LEN;
 	rec->session.iscsi.MaxBurstLength = DEF_INI_MAX_BURST_LEN;
-	rec->session.iscsi.DefaultTime2Wait = 0;
+	rec->session.iscsi.DefaultTime2Wait = ISCSI_DEF_TIME2WAIT;
 	rec->session.iscsi.DefaultTime2Retain = 0;
 	rec->session.iscsi.MaxConnections = 1;
 	rec->session.iscsi.MaxOutstandingR2T = 1;
 	rec->session.iscsi.ERL = 0;
+	rec->session.iscsi.FastAbort = 1;
 
 	for (i=0; i<ISCSI_CONN_MAX; i++) {
-		rec->conn[i].startup = 0;
+		rec->conn[i].startup = ISCSI_STARTUP_MANUAL;
 		rec->conn[i].port = ISCSI_LISTEN_PORT;
 		rec->conn[i].tcp.window_size = TCP_WINDOW_SIZE;
 		rec->conn[i].tcp.type_of_service = 0;
 		rec->conn[i].timeo.login_timeout= DEF_LOGIN_TIMEO;
 		rec->conn[i].timeo.logout_timeout= DEF_LOGOUT_TIMEO;
 		rec->conn[i].timeo.auth_timeout = 45;
-		rec->conn[i].timeo.active_timeout=5;
-		rec->conn[i].timeo.idle_timeout = 60;
-		rec->conn[i].timeo.ping_timeout = 5;
 
 		rec->conn[i].timeo.noop_out_interval = DEF_NOOP_OUT_INTERVAL;
 		rec->conn[i].timeo.noop_out_timeout = DEF_NOOP_OUT_TIMEO;
 
 		rec->conn[i].iscsi.MaxRecvDataSegmentLength =
 						DEF_INI_MAX_RECV_SEG_LEN;
-		rec->conn[i].iscsi.HeaderDigest = CONFIG_DIGEST_PREFER_OFF;
+		rec->conn[i].iscsi.HeaderDigest = CONFIG_DIGEST_NEVER;
 		rec->conn[i].iscsi.DataDigest = CONFIG_DIGEST_NEVER;
 		rec->conn[i].iscsi.IFMarker = 0;
 		rec->conn[i].iscsi.OFMarker = 0;
 	}
 
-	iface_init(&rec->iface);
+	iface_setup_defaults(&rec->iface);
 }
 
 void iscsid_handle_error(mgmt_ipc_err_e err)
@@ -252,6 +330,7 @@ void iscsid_handle_error(mgmt_ipc_err_e err)
 		/* 16 */ "Unknown request",
 		/* 17 */ "encountered iSNS failure",
 		/* 18 */ "could not communicate to iscsid",
+		/* 19 */ "encountered non-retryable iSCSI login failure",
 	};
 	log_error("initiator reported error (%d - %s)", err, err_msgs[err]);
 }
@@ -286,11 +365,7 @@ int __iscsi_match_session(node_rec_t *rec, char *targetname,
 	if (rec->conn[0].port != -1 && port != rec->conn[0].port)
 		return 0;
 
-	if (strlen(rec->iface.transport_name) &&
-	    strcmp(rec->iface.transport_name, iface->transport_name))
-		return 0;
-
-	if (!iface_match_bind_info(&rec->iface, iface))
+	if (!iface_match(&rec->iface, iface))
 		return 0;
 
 	return 1;
